#
# NES screen builder
# Brad Smith 2/12/2013
# http://rainwarrior.ca
#
# Python 3
#

import array
import tkinter
import struct
import subprocess

SPRITE_H = 8
#SPRITE_H = 16

#
# image class
#

# simple 24-bit bitmap class
class Bitmap:

    # converts tkinter.PhotoImage to Bitmap
    def LoadPhotoImage(self,img):
        self.w = img.width()
        self.h = img.height()
        size = self.w * self.h
        self.data = array.array('I',(0 for i in range(0,size)))
        for y in range(0,self.h):
            for x in range(0,self.w):
                (r,g,b) = img.get(x,y).split()
                p = (int(r) << 16) | (int(g) << 8) | int(b)
                self.Set(x,y,p)

    # converts Bitmap into tkinter.Photoimage
    def MakePhotoImage(self):
        pi = tkinter.PhotoImage(width=self.w, height=self.h)
        for y in range(0,self.h):
            for x in range(0,self.w):
                pi.put("#%06x" % (self.Get(x,y)), (x,y))
        return pi

    # loads GIF into Bitmap
    def LoadGIF(self,filename):
        img = tkinter.PhotoImage(file=filename)
        self.LoadPhotoImage(img)

    # saves a TGA
    def SaveTGA(self,filename):
        file = open(filename, mode="wb")
        file.write(struct.pack("<BBBHHBHHHHBB",
                               0, # ID field (none)
                               0, # no colour map
                               2, # RGB
                               1, 2, 3, # no colour map
                               4, 5, # origin
                               self.w,
                               self.h,
                               24, # BPP
                               0x20, # V flip, no alpha
                               ))
        for y in range(0,self.h):
            for x in range(0,self.w):
                p = self.Get(x,y)
                file.write(struct.pack("<BBB",
                                       p & 0xFF,
                                       (p >> 8) & 0xFF,
                                       (p >> 16) & 0xFF))
        file.close()

    # gets a pixel
    def Get(self,x,y):
        return self.data[x+(y*self.w)]

    # sets a pixel
    def Set(self,x,y,p):
        self.data[x+(y*self.w)] = p

    # new image of specified dimension
    def New(self,w,h):
        self.w = w
        self.h = h
        size = self.w * self.h
        self.data = array.array('I',(0 for i in range(0,size)))

    # copy from image
    def Copy(self,img):
        self.w = img.w
        self.h = img.h
        self.data = array.array('I',img.data)

    # copy region from image
    def CopySub(self,img,x,y,w,h):
        self.New(w,h)
        for oy in range(0,h):
            for ox in range(0,w):
                self.Set(ox,oy,img.Get(x+ox,y+oy))

    # compare images
    def Same(self,img):
        if self.w != img.w or self.h != img.h:
            return False
        for y in range(0,self.h):
            for x in range(0,self.w):
                if self.Get(x,y) != img.Get(x,y):
                    return False
        return True

    # flip self
    def FlipH(self):
        img = Bitmap()
        img.New(self.w, self.h)
        for y in range(0,self.h):
            for x in range(0,self.w):
                img.Set(x,y,self.Get((self.w-1)-x,y))
        return img

    # flip self
    def FlipV(self):
        img = Bitmap()
        img.New(self.w, self.h)
        for y in range(0,self.h):
            for x in range(0,self.w):
                img.Set(x,y,self.Get(x,(self.h-1)-y))
        return img

    # blit image onto self
    def Blit(self, img, x, y, bgcolor):
        for ay in range(0,img.h):
            for ax in range(0,img.w):
                dx = ax + x
                dy = ay + y
                if (dx >= 0 and dx < self.w and dy >= 0 and dy < self.h):
                    p = img.Get(ax,ay)
                    if (p != bgcolor):
                        self.Set(dx,dy,img.Get(ax,ay))

#
# image manipulation
#

# list colors in image
def list_colors(img):
    c = []
    for y in range(0,img.h):
        for x in range(0,img.w):
            p = img.Get(x,y)
            if p not in c:
                c.append(p)
    count = 0
    for p in c:
        print("%3d: %06X" % (count,p))
        count += 1


# count pixels of a colour
def count_color(img, col):
    count = 0
    for y in range(0,img.h):
        for x in range(0,img.w):
            if (img.Get(x,y) == col):
                count += 1
    return count

# replace one colour with another
def replace_color(img, cola, colb):
    for y in range(0,img.h):
        for x in range(0,img.w):
            if (img.Get(x,y) == cola):
                img.Set(x,y,colb)

# normalizes an image to 256-colour greyscale
def normalize(img):
    cmin = 0x256;
    cmax = 0x0
    for y in range(0,img.h):
        for x in range(0,img.w):
            p = img.Get(x,y)
            if (p < cmin):
                cmin = p
            if (p > cmax):
                cmax = p
    for y in range(0,img.h):
        for x in range(0,img.w):
            p = img.Get(x,y)
            p = int((p-cmin) * 255 / (cmax-cmin))
            img.Set(x,y,p + (p<<8) + (p<<16))

# remap colours in image
def remap_image(img, colormap):
    for y in range(0,img.h):
        for x in range(0,img.w):
            p = img.Get(x,y)
            p = colormap[p]
            img.Set(x,y,p)

# reverse a colourmap (uses first found if two map to same colour)
def reverse_map(m):
    mo = {}
    for (a,b) in m.items():
        if b not in mo:
            mo[b] = a
    return mo

# collects pixels in a block
def collect_pixels(img, x, y, w, h):
    p = []
    for iy in range(y,y+h):
        for ix in range(x,x+w):
            p.append(img.Get(ix,iy))
    return p

# finds best pal among list of pals to apply to pixels
def best_pal(pixels, pals):
    c = array.array('I', (0 for i in range(0,len(pals))))
    for pi in range(0,len(pals)):
        pal = pals[pi]
        count = 0
        for p in pixels:
            if p in pal:
                count += 1
        c[pi] = count
    bp = 0
    bc = 0
    for i in range(0,len(c)):
        if c[i] > bc:
            bc = c[i]
            bp = i
    return bp

# eliminates out-of-palette pixels in specified block (replaces with pal[0])
def palette_block(img,ax,ay,aw,ah,pal):
    for y in range(ay,ay+ah):
        for x in range(ax,ax+aw):
            p = img.Get(x,y)
            if p not in pal:
                img.Set(x,y,pal[0])

#
# NES spacific image manipulation
#
           
# generate "leftover" image
def leftover(target, current, bgcolor):
    img = Bitmap()
    img.Copy(target)
    for y in range(0,img.h):
        for x in range(0,img.w):
            if current.Get(x,y) == target.Get(x,y):
                img.Set(x,y,bgcolor)
    return img

# find best sprite to subtract from leftover image
def target_sprite(img, pals):
    bgcolor = pals[0][0]
    bcnt = 0
    bpos = None
    # brute force search of all sprite locations, very slow
    yskip = 0
    for y in range(0,img.h-SPRITE_H):
        if (yskip > 0):
            yskip -= 1
            continue
        ybcnt = 0 # count best on column for speedup
        xskip = 0
        for x in range(0,img.w-8):
            if (xskip > 0):
                xskip -= 1
                continue
            xbcnt = 0 # count best on row for speedup
            xlow = False # make sure leftmost aligned
            ylow = False # make sure topmost aligned
            for tp in range(0,4):
                count = 0
                pal = pals[tp]
                for ty in range(0,SPRITE_H):
                    for tx in range(0,8):
                        p = img.Get(x+tx,y+ty)
                        if (p != bgcolor and p in pal):
                            count += 1
                if (count > bcnt):
                    bcnt = count
                    bpos = (x,y)
                if (count > xbcnt):
                    xbcnt = count
                if (count > ybcnt):
                    ybcnt = count
            if (xbcnt == 0):
                xskip = 7 # skip ahead if this block was empty
        if (ybcnt == 0):
            yskip = SPRITE_H-1 # skip ahead if this row of blocks was empty
    return bpos

# global count of sprites
sprite_count = 0

# remove a sprite from the image (using whatever palette takes the most pixels from the block)
def remove_sprite(img,sx,sy,pals):
    global sprite_count
    bgcolor = pals[0][0]
    p = collect_pixels(img,sx,sy,8,SPRITE_H)
    pi = best_pal(p,pals)
    pal = pals[pi]
    count = 0
    spr = Bitmap()
    spr.New(8,SPRITE_H)
    replace_color(spr,0,bgcolor)
    for y in range(0,SPRITE_H):
        for x in range(0,8):
            p = img.Get(sx+x,sy+y)
            if p != bgcolor:
                if p in pal:
                    img.Set(sx+x,sy+y,bgcolor)
                    spr.Set(x,y,p)
                    count += 1
    print("Sprite %3d: %3d, %3d pal %d, pixels: %d" % (sprite_count,sx,sy,pi,count))
    sprite_count += 1
    return spr

# paint col in sprites' region
def show_sprite(img_in,sx,sy,bgcol,col):
    img = Bitmap()
    img.Copy(img_in)
    for y in range(0,SPRITE_H):
        for x in range(0,8):
            p = img.Get(sx+x,sy+y)
            if (p == bgcol):
                img.Set(sx+x,sy+y,col)
    return img

# increment pixel in sprites' region
def mark_sprite(img,sx,sy):
    for y in range(0,SPRITE_H):
        for x in range(0,8):
            p = img.Get(sx+x,sy+y)
            img.Set(sx+x,sy+y,p+1)

# attributify a nametable
def nametable_gen(img, pals):
    for ay in range(0,int(img.h / 16)):
        for ax in range(0,int(img.w / 16)):
            p = collect_pixels(img, ax*16, ay*16, 16, 16)
            pal = pals[best_pal(p,pals)]
            palette_block(img, ax*16, ay*16, 16, 16, pal)

# pattern builder
def build_pat(pats,bgcolor):
    pat = Bitmap()
    pat.New(128,128)
    replace_color(pat,0,bgcolor)
    for i in range(0,len(pats)):
        pat.Blit(pats[i],(i&15)*8,int(i/16)*pats[0].h,-1)
    return pat

# chr generator
def nametable_pat(img, bgcolor):
    pats = []
    grid = array.array('I',(0 for i in range(0,int(img.w/8)*int(img.h/8))))
    gi = 0
    for ty in range(0,int(img.h / 8)):
        for tx in range(0,int(img.w / 8)):
            tile = Bitmap()
            tile.CopySub(img,tx*8,ty*8,8,8)
            duplicate = -1
            # check if this tile already exists
            for i in range(0,len(pats)):
                if pats[i].Same(tile):
                    duplicate = i
                    break
            if duplicate == -1:
                duplicate = len(pats)
                pats.append(tile)
            # save nametable tile
            grid[gi] = duplicate
            gi += 1
    return (build_pat(pats,bgcolor), grid)

# takes a nametable from nametable_gen and reduces to 4-colour indexed + attribute data
def build_att(img, pals):
    ind = Bitmap()
    ind.New(img.w, img.h)
    att = array.array('I',(0 for i in range(0,int(img.h/16)*int(img.w/16))))
    ai = 0 # index to att
    for ay in range(0,int(img.h/16)):
        for ax in range(0,int(img.w/16)):
            # find palette for this block
            p = collect_pixels(img,ax*16,ay*16,16,16)
            pi = best_pal(p,pals)
            pal = pals[pi]
            # store attribute
            att[ai] = pi
            ai += 1
            # index colours
            for y in range(0,16):
                for x in range(0,16):
                    p = img.Get(x+(ax*16),y+(ay*16))
                    i = 0
                    # find matching colour
                    for pc in range(0,len(pal)):
                        if p == pal[pc]:
                            i = pc
                            break
                    # store
                    ind.Set(x+(ax*16),y+(ay*16),i)
    return (ind, att)

# takes a list of sprites and positions and palettes,
# builds indexed sprite pattern table, and OAM data
def build_oam(sprites,pos,pals):
    oam = array.array('I',(0 for i in range(0,256)))
    # index sprite images and build OAM data
    for s in range(0,len(sprites)):
        sprite = sprites[s]
        sx = pos[s][0]
        sy = pos[s][1]
        pi = best_pal(collect_pixels(sprite,0,0,8,SPRITE_H),pals)
        # save OAM
        oam[(4*s)+0] = sy - 1
        oam[(4*s)+1] = s
        oam[(4*s)+2] = pi
        oam[(4*s)+3] = sx
        # index the sprite
        pal = pals[pi]
        for y in range(0,sprite.h):
            for x in range(0,sprite.h):
                p = sprite.Get(x,y)
                i = 0
                # find matching colour
                for pc in range(0,len(pal)):
                    if p == pal[pc]:
                        i = pc
                        break
                # store
                sprite.Set(x,y,i)
    # hide unused sprites
    for s in range(len(sprites),64):
        oam[(4*s)+0] = 0xFF
        oam[(4*s)+1] = 0xFF
        oam[(4*s)+2] = 0xFF
        oam[(4*s)+3] = 0xFF
    # build indexed pattern
    pat = build_pat(sprites,0)
    return (pat, oam)

#
# NES file packing
#

# converts indexed pattern image into PPU planar data
def save_pat(pat, filename):
    file = open(filename, mode="wb")
    for ty in range(0,int(pat.h / 8)):
        for tx in range(0,int(pat.w / 8)):
            sx = tx * 8
            sy = ty * 8
            tile = array.array('B',(0 for i in range(0,16)))
            for line in range(0,8):
                bitmask = 0x80
                for x in range(0,8):
                    p = pat.Get(x+sx,line+sy)
                    if ((p & 1) == 1):
                        tile[line] |= bitmask
                    if ((p & 2) == 2):
                        tile[line+8] |= bitmask
                    bitmask >>= 1
            file.write(tile)

# coverts 16x15 attribute list into packed PPU attribute table
def save_att(att, filename):
    # helper function for reading attribute
    def read_att(att,x,y):
        if (y >= 15):
            return 0
        else:
            return att[(y*16)+x]
    # actual function
    file = open(filename, mode="wb")
    for a in range(0,64):
        # top left coordinate of quad
        tx = (a & 7) * 2
        ty = int(a / 8) * 2
        b = (read_att(att,tx+0,ty+0) << 0) | \
            (read_att(att,tx+1,ty+0) << 2) | \
            (read_att(att,tx+0,ty+1) << 4) | \
            (read_att(att,tx+1,ty+1) << 6)
        file.write(struct.pack("<B",b))

# writes palettes to file
def save_pals(pals, filename):
    file = open(filename, mode="wb")
    for pal in pals:
        for p in pal:
            file.write(struct.pack("<B",p))

# writes blk to file, packing each element as a byte
def save_bytes(blk, filename):
    file = open(filename, mode="wb")
    for b in blk:
        file.write(struct.pack("<B",b))

#
# specific stuff, main program
#

PAL_BG = 0x1B # background green
PAL_BS = 0x0B # background shadow
PAL_BL = 0x0F # text black, murray black
PAL_WH = 0x30 # text white
PAL_GR = 0x10 # light grey
PAL_SL = 0x36 # skin light
PAL_SM = 0x26 # skin mid
PAL_SD = 0x16 # skin dark (maybe 0x17?)
PAL_SS = 0x07 # coat/skin shadow
PAL_SP = 0xFF # sprite marker

pal_nmt = [
    # background palettes
    [ PAL_BG, PAL_BL, PAL_BS, PAL_WH ],
    [ PAL_BG, PAL_SM, PAL_SD, PAL_SL ],
    [ PAL_BG, PAL_BL, PAL_SD, PAL_SS ], #[ PAL_BG, PAL_BL, PAL_GR, PAL_SS ],
    [ PAL_BG, PAL_BL, PAL_BS, PAL_SS ],
    ]

pal_spr = [
    # sprite palettes
    [ PAL_BG, PAL_BL, PAL_SS, PAL_SD ], #[ PAL_BG, PAL_BL, PAL_SS, PAL_GR ],
    [ PAL_BG, PAL_SD, PAL_SM, PAL_SL ],
    [ PAL_BG, PAL_BS, PAL_BL, PAL_SS ], #[ PAL_BG, PAL_BS, PAL_BL, PAL_GR ],
    [ PAL_BG, PAL_BL, PAL_SS, PAL_WH ],
    ]

palmap = {
    0x04AA4C : PAL_BG, # background green
    0x045624 : PAL_BS, # background shadow
    0x040204 : PAL_BL, # text black
    0xFCFAFC : PAL_WH, # text white
    0xECC29C : PAL_SL, # skin light
    0xD47E54 : PAL_SM, # skin mid
    0xA44224 : PAL_SD, # skin dark
    0x542A1C : PAL_SS, # coat/skin shadow
    0x0C0A04 : PAL_BL, # murray black
    0x8C867C : PAL_GR, # light grey
    0xFF0000 : PAL_SP, # sprite marker
    }

pamlap = reverse_map(palmap)

def ShowImage(tk, b):
    pi = b.MakePhotoImage()
    l = tkinter.Label(tk, image=pi)
    l.pack()
    tk.mainloop()
    tk = tkinter.Tk()

def ShowImageNES(tk, b):
    img = Bitmap()
    img.Copy(b)
    remap_image(img,pamlap)
    ShowImage(tk,img)

def SaveNESTGA(b, filename):
    img = Bitmap()
    img.Copy(b)
    remap_image(img,pamlap)
    img.SaveTGA(filename)

def SaveNrmTGA(b, filename):
    img = Bitmap()
    img.Copy(b)
    normalize(img)
    img.SaveTGA(filename)

if __name__=="__main__":
    subprocess.call("del bill\\*.tga", shell=True)
    subprocess.call("del bill\\*.bin", shell=True)
    tk = tkinter.Tk()
    bgcolor = pal_nmt[0][0]
    # load image
    a = Bitmap()
    a.LoadGIF("billmurray.gif")
    # remap to NES palette
    remap_image(a,palmap)
    SaveNESTGA(a,"bill\\begin.tga")
    print("Colors in image:")
    list_colors(a)
    # attributify
    nmt = Bitmap()
    nmt.Copy(a)
    nametable_gen(nmt,pal_nmt)
    SaveNESTGA(nmt,"bill\\nmt_color.tga")
    print("Colors in nametable:")
    list_colors(nmt)
    # nametable patterns
    (pat, grid) = nametable_pat(nmt,bgcolor)
    SaveNESTGA(pat,"bill\\pat0_color.tga")
    # build attributes and indexed colour
    (nmt_ind, att) = build_att(nmt,pal_nmt)
    SaveNrmTGA(nmt_ind,"bill\\nmt_index.tga")
    (pat, grad) = nametable_pat(nmt_ind,0)
    SaveNrmTGA(pat,"bill\\pat0_index.tga")
    pat.SaveTGA("bill\\pat0_index_test.tga")
    # save grid + att as nametable binary, and pat0
    save_pat(pat, "bill\\pat0.bin")
    save_bytes(grid, "bill\\grid.bin")
    save_att(att, "bill\\att.bin")
    save_pals(pal_nmt, "bill\\pal0.bin")
    # what's left? cover it with sprites
    lft = leftover(a,nmt,bgcolor)
    SaveNESTGA(lft,"bill\\sprite000.tga")
    print("Leftover colors for sprites:")
    list_colors(lft)
    lft_pix = (lft.w*lft.h) - count_color(lft,bgcolor)
    print("Leftover pixels: %d (opt %d sprites)" % (lft_pix, lft_pix/(8*SPRITE_H)))
    # gradually remove sprites one by one
    smp = Bitmap()
    smp.New(lft.w,lft.h)
    target = target_sprite(lft,pal_spr)
    sprites = []
    sprite_pos = []
    while None != target:
        sprites.append(remove_sprite(lft,target[0],target[1],pal_spr))
        sprite_pos.append((target[0],target[1]))
        SaveNESTGA(show_sprite(lft,target[0],target[1],bgcolor,PAL_SP)
                   ,"bill\\sprite%03d.tga" % (sprite_count))
        mark_sprite(smp,target[0],target[1])
        target = target_sprite(lft,pal_spr)
    # check for overlap
    for i in range(0,240):
        count = 0
        for (x,y) in sprite_pos:
            if (y <= i) and ((y + SPRITE_H - 1) >= i):
                count += 1
        print("Scanline %3d: %d sprites" % (i, count))
        if (count > 8):
            print("Too many sprites on line: %d" % (i))
    # show sprites
    SaveNESTGA(build_pat(sprites,bgcolor),"bill\\pat1_color.tga")
    SaveNrmTGA(smp,"bill\\sprite_map.tga")
    # build indexed sprite pattern and OAM
    (pat, oam) = build_oam(sprites,sprite_pos,pal_spr)
    SaveNrmTGA(pat,"bill\\pat1_index.tga")
    # save OAM, save PAT1
    save_pat(pat, "bill\\pat1.bin")
    save_bytes(oam, "bill\\oam.bin")
    save_pals(pal_spr, "bill\\pal1.bin")
    # done
    SaveNESTGA(lft,"bill\\final_left.tga")
    #ShowImageNES(tk, lft)
